
# Aquilify devops / Test _CSP @noql :: 3217

import json
import logging
import secrets
import hashlib
from pathlib import Path
from typing import Dict, Union, List, Optional

from ..wrappers import Request, Response
from ..exception.__handler import handle_exception

class CSPMiddleware:
    def __init__(
        self,
        csp_directives: Dict[str, Union[str, List[str]]],
        report_uri: str = '',
        report_only: bool = False,
        enable_violation_handling: bool = False,
        report_sample_weight: float = 0.0,
        violation_report_endpoint: str = '/violation-report',
        log_file_path: str = 'csp_violation_reports.log',
        security_headers: Dict[str, str] = None,
        nonce_length: int = 16,
        nonce_algorithm: str = 'sha256',
        referrer_policy: str = 'strict-origin-when-cross-origin',
        hsts_max_age: int = 31536000,
        hsts_include_subdomains: bool = True,
        hsts_preload: bool = False,
        feature_policy: Optional[Dict[str, str]] = None,
        x_content_type_options: str = 'nosniff',
        x_frame_options: str = 'DENY',
        x_xss_protection: str = '1; mode=block',
        expect_ct: Optional[str] = None,
        cross_origin_opener_policy: Optional[str] = None,
        cross_origin_embedder_policy: Optional[str] = None,
        force_https: bool = False,
        referrer_policy_feature: bool = False,
        referrer_policy_no_referer: bool = False,
        referrer_policy_no_referrer_when_downgrade: bool = False,
    ) -> None:
        self.csp_directives = csp_directives
        self.report_uri = report_uri
        self.report_only = report_only
        self.enable_violation_handling = enable_violation_handling
        self.report_sample_weight = report_sample_weight
        self.violation_report_endpoint = violation_report_endpoint
        self.log_file_path = Path(log_file_path)
        self.security_headers = security_headers or {}
        self._setup_violation_log_file()
        self.logger = self._setup_logging()
        self.violation_reports = [],
        self.nonce_length: int = nonce_length
        self.nonce_algorithm: str = nonce_algorithm
        self.referrer_policy: str = referrer_policy
        self.hsts_max_age: int = hsts_max_age
        self.hsts_include_subdomains: bool = hsts_include_subdomains
        self.hsts_preload: bool = hsts_preload
        self.feature_policy: Dict[str, str] = feature_policy if feature_policy else {}
        self.x_content_type_options: str = x_content_type_options
        self.x_frame_options: str = x_frame_options
        self.x_xss_protection: str = x_xss_protection
        self.expect_ct: Optional[str] = expect_ct
        self.cross_origin_opener_policy: Optional[str] = cross_origin_opener_policy
        self.cross_origin_embedder_policy: Optional[str] = cross_origin_embedder_policy
        self.force_https: bool = force_https
        self.referrer_policy_feature: bool = referrer_policy_feature
        self.referrer_policy_no_referer: bool = referrer_policy_no_referer
        self.referrer_policy_no_referrer_when_downgrade: bool = referrer_policy_no_referrer_when_downgrade

    async def __call__(self, request: Request, response: Response) -> Response:
        try:
            if self.force_https and await request.scheme() != "https":
                    response.status_code = 400
                    response.content = "<h1>400 | Bad Request - HTTPS Required!</h1>"
                    response.content_type = 'text/html'
                    return response
            
            if self._is_violation_report_request(request):
                await self._handle_violation_reports(request)
                return Response(content='Violation report received and logged.', status_code=200)
            response = self._set_csp_header(response)
            response = self._set_additional_security_headers(response)

            if self.expect_ct:
                response['headers']["Expect-CT"] = self.expect_ct

            if self.cross_origin_opener_policy:
                response['headers']["Cross-Origin-Opener-Policy"] = self.cross_origin_opener_policy

            if self.cross_origin_embedder_policy:
                response['headers']["Cross-Origin-Embedder-Policy"] = self.cross_origin_embedder_policy

            if self.referrer_policy_feature:
                response.headers["Referrer-Policy"] = self.referrer_policy + ", " + "features"

            if self.referrer_policy_no_referer:
                response.headers["Referrer-Policy"] = self.referrer_policy + ", " + "no-referer"

            if self.referrer_policy_no_referrer_when_downgrade:
                response.headers["Referrer-Policy"] = self.referrer_policy + ", " + "no-referrer-when-downgrade"

            response.headers["Referrer-Policy"] = self.referrer_policy
            hsts_directive = f"max-age={self.hsts_max_age}"
            if self.hsts_include_subdomains:
                hsts_directive += "; includeSubDomains"
            if self.hsts_preload:
                hsts_directive += "; preload"
            response.headers["Strict-Transport-Security"] = hsts_directive

            if self.feature_policy:
                feature_policy_directives = " ".join([f"{key} {value}" for key, value in self.feature_policy.items()])
                response.headers["Feature-Policy"] = feature_policy_directives

            response.headers["X-Content-Type-Options"] = self.x_content_type_options
            response.headers["X-Frame-Options"] = self.x_frame_options
            response.headers["X-XSS-Protection"] = self.x_xss_protection
            return response
        except Exception as e:
            await handle_exception(e)
        return response

    def _is_violation_report_request(self, request: Request) -> bool:
        return request.scope['path'] == self.violation_report_endpoint and request.method == 'POST'

    def _set_csp_header(self, response: Response) -> Response:
        csp_header_value = self._generate_csp_header()
        response.headers['Content-Security-Policy' if not self.report_only else 'Content-Security-Policy-Report-Only'] = csp_header_value
        if self.report_uri:
            response.headers['Content-Security-Policy-Report-Only'] = f"{csp_header_value}; report-uri {self.report_uri}"
        return response

    def _generate_csp_header(self) -> str:
        csp_directive_strings = [f"{directive} {self._generate_directive_value(directive)}" for directive in self.csp_directives.keys()]
        return "; ".join(csp_directive_strings)

    def _generate_directive_value(self, directive: str) -> str:
        value = self.csp_directives[directive]
        return self._join_values(self._generate_nonce_for_inline(value) if self._is_inline_directive(directive) else value)

    def _generate_nonce_for_inline(self, directives: Union[str, List[str]]) -> Union[str, List[str]]:
        return [self._generate_nonce() if "'nonce'" in directive else directive for directive in directives] if isinstance(directives, list) else self._generate_nonce() if "'nonce'" in directives else directives

    def _generate_nonce(self) -> str:
        nonce = secrets.token_urlsafe(self.nonce_length)
        if self.nonce_algorithm:
            hash_func = getattr(hashlib, self.nonce_algorithm, hashlib.sha256)
            hashed_nonce = hash_func(nonce.encode()).digest()
            return hashed_nonce.hex()
        return nonce

    def _is_inline_directive(self, directive: str) -> bool:
        return directive in ['script-src', 'style-src']

    def _join_values(self, value: Union[str, List[str]]) -> str:
        return " ".join(value) if isinstance(value, list) else value

    def _setup_violation_log_file(self) -> None:
        if not self.log_file_path.is_file():
            self.log_file_path.touch()

    def _setup_logging(self) -> logging.Logger:
        logger = logging.getLogger('CSPViolationLogger')
        logger.setLevel(logging.INFO)
        handler = logging.FileHandler(self.log_file_path)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return logger

    async def _handle_violation_reports(self, request: Request) -> None:
        violation_report = await request.json()
        self._store_violation_report(violation_report)
        self.logger.info(json.dumps(violation_report))
        self.violation_reports.append(violation_report)
        if self.enable_violation_handling:
            self._process_violation_reports()

    def _store_violation_report(self, violation_report: dict) -> None:
        with self.log_file_path.open(mode='a') as file:
            file.write(json.dumps(violation_report) + '\n')

    def _process_violation_reports(self) -> None:
        num_reports = len(self.violation_reports)
        if num_reports > 0:
            weighted_reports = int(self.report_sample_weight * num_reports)
            reports_to_process = self.violation_reports[:weighted_reports]

    def _set_additional_security_headers(self, response: Response) -> Response:
        for header, value in self.security_headers.items():
            response.headers[header] = value
        return response



import os
import asyncio
import mimetypes

class FileStorage:
    def __init__(self, filename, content_type, data):
        self.filename = filename
        self.content_type = content_type
        self.data = data

    @property
    def content_length(self):
        return len(self.data)

    @property
    def mimetype(self):
        return self.content_type

    @property
    def name(self):
        return self.filename

    @classmethod
    async def create(cls, filename, content_type, data):
        return cls(filename, content_type, data)

    async def save(self, directory, overwrite=False):
        # Create the directory if it doesn't exist
        os.makedirs(directory, exist_ok=True)

        file_path = os.path.join(directory, self.filename)
        if not overwrite:
            # If the file already exists, add a suffix to the filename to make it unique
            file_name, file_extension = os.path.splitext(self.filename)
            count = 1
            while os.path.exists(file_path):
                new_filename = f"{file_name}_{count}{file_extension}"
                file_path = os.path.join(directory, new_filename)
                count += 1

        with open(file_path, 'wb') as f:
            f.write(self.data)

        return file_path  # Return the file path where the file is saved

    def get_extension(self):
        _, extension = os.path.splitext(self.filename)
        return extension

    def is_image(self):
        return self.mimetype.startswith('image/')

    def is_text(self):
        return self.mimetype.startswith('text/')

    def guess_extension(self):
        return mimetypes.guess_extension(self.mimetype)

    def guess_type(self):
        return mimetypes.guess_type(self.filename)


import json
import zlib
import gzip
import time
import base64
import hashlib
import random
import logging
import platform
import http.client

from urllib.parse import urlparse, urljoin
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Union, Dict, Any, List

from .response import Response

class HttpException(Exception):
    def __init__(self, status_code: int, reason: str = None):
        self.status_code = status_code
        self.reason = reason
        super().__init__(f"HTTP {status_code} - {reason}")

    def __str__(self) -> str:
        error_message = f"HTTP {self.status_code} - {self.reason}\n"
        return error_message

class ClientSession:
    def __init__(self):
        self.default_timeout: int = 10 
        self.max_redirects: int = 5
        self.session: Optional[Any] = None
        self.connection_pool: Dict[str, http.client.HTTPConnection] = {}
        self.user_agents: List[str] = self._get_user_agent()
        self.logger: logging.Logger = self._configure_logger()
        self.executor: ThreadPoolExecutor = ThreadPoolExecutor(max_workers=10)
        self.request_retries: int = 0
        self.cache: Dict[str, Any] = {}
        self.response_cache_max_size: int = 1000
        self.default_headers: Dict[str, str] = {} 
        self.proxy: Optional[Dict[str, Union[str, int]]] = None 
        self.use_response_cache: bool = True

    def _get_user_agent(self) -> List[str]:
        user_agent = f"Python/{platform.python_version()} ({platform.system()}; {platform.release()})"
        return [user_agent]

    def _configure_logger(self) -> logging.Logger:
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch = logging.StreamHandler()
        ch.setFormatter(formatter)
        logger.addHandler(ch)
        return logger

    def close_connections(self) -> None:
        for conn in self.connection_pool.values():
            conn.close()
        self.connection_pool.clear()

    def get(self, url: str, headers: Optional[Dict[str, str]] = None, params: Optional[Dict[str, Union[str, int]]] = None, 
            timeout: Optional[float] = None, retries: Optional[int] = None, use_cache: bool = True, auth: Optional[Dict[str, str]] = None) -> Response:
        return self.request("GET", url, headers=headers, params=params, timeout=timeout, retries=retries, use_cache=use_cache, auth=auth)

    def post(self, url: str, headers: Optional[Dict[str, str]] = None, data: Optional[str] = None, json_data: Optional[Dict[str, Any]] = None,
             files: Optional[Dict[str, Any]] = None, timeout: Optional[float] = None, retries: Optional[int] = None, use_cache: bool = True, 
             auth: Optional[Dict[str, str]] = None) -> Response:
        return self.request("POST", url, headers=headers, data=data, json_data=json_data, files=files, timeout=timeout, retries=retries, use_cache=use_cache, auth=auth)

    def put(self, url: str, headers: Optional[Dict[str, str]] = None, data: Optional[str] = None, json_data: Optional[Dict[str, Any]] = None,
            timeout: Optional[float] = None, retries: Optional[int] = None, use_cache: bool = True, auth: Optional[Dict[str, str]] = None) -> Response:
        return self.request("PUT", url, headers=headers, data=data, json_data=json_data, timeout=timeout, retries=retries, use_cache=use_cache, auth=auth)

    def delete(self, url: str, headers: Optional[Dict[str, str]] = None, timeout: Optional[float] = None, retries: Optional[int] = None, use_cache: bool = True, auth: Optional[Dict[str, str]] = None) -> Response:
        return self.request("DELETE", url, headers=headers, timeout=timeout, retries=retries, use_cache=use_cache, auth=auth)

    def head(self, url: str, headers: Optional[Dict[str, str]] = None, params: Optional[Dict[str, Union[str, int]]] = None, 
             timeout: Optional[float] = None, retries: Optional[int] = None, use_cache: bool = True, auth: Optional[Dict[str, str]] = None) -> Response:
        return self.request("HEAD", url, headers=headers, params=params, timeout=timeout, retries=retries, use_cache=use_cache, auth=auth)

    def patch(self, url: str, headers: Optional[Dict[str, str]] = None, data: Optional[str] = None, json_data: Optional[Dict[str, Any]] = None,
              timeout: Optional[float] = None, retries: Optional[int] = None, use_cache: bool = True, auth: Optional[Dict[str, str]] = None) -> Response:
        return self.request("PATCH", url, headers=headers, data=data, json_data=json_data, timeout=timeout, retries=retries, use_cache=use_cache, auth=auth)

    def options(self, url: str, headers: Optional[Dict[str, str]] = None, params: Optional[Dict[str, Union[str, int]]] = None, 
                timeout: Optional[float] = None, retries: Optional[int] = None, use_cache: bool = True, auth: Optional[Dict[str, str]] = None) -> Response:
        return self.request("OPTIONS", url, headers=headers, params=params, timeout=timeout, retries=retries, use_cache=use_cache, auth=auth)

    def request(
        self,
        method: str,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, Union[str, int]]] = None,
        data: Optional[str] = None,
        json_data: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        use_cache: bool = True,
        auth: Optional[Dict[str, str]] = None
    ) -> Optional[Response]:
        if timeout is None:
            timeout = self.default_timeout
        if retries is None:
            retries = self.request_retries

        cache_key = self._generate_cache_key(method, url, headers, params, data, json_data)
        if use_cache and cache_key in self.cache:
            return self.cache[cache_key]

        for _ in range(retries + 1):
            try:
                response = self._make_request(method, url, headers, params, data, json_data, files, timeout, auth)
                if response is not None:
                    if use_cache:
                        self._cache_response(cache_key, response)
                    return response
            except HttpException as e:
                self.logger.error(f"Request error: {e}")
                time.sleep(2 ** _)  # Exponential backoff for retries
            except Exception as e:
                self.logger.error(f"Request error: {type(e).__name__}: {e}")
        return None

    def _make_request(
        self,
        method: str,
        url: str,
        headers: Optional[Dict[str, str]],
        params: Optional[Dict[str, Union[str, int]]],
        data: Optional[str],
        json_data: Optional[Dict[str, Any]],
        files: Optional[Dict[str, Any]],
        timeout: Optional[float],
        auth: Optional[Dict[str, str]]
    ) -> Optional[Response]:
        redirect_count = 0

        while redirect_count < self.max_redirects:
            # Parse the URL
            url_parsed = urlparse(url)
            path = url_parsed.path or '/'
            query = url_parsed.query

            # Extract the host
            host = url_parsed.hostname
            if host is None:
                raise ValueError("Invalid URL. Unable to extract hostname.")

            # Extract the port
            port = url_parsed.port or 80

            conn_key = f"{host}:{port}"
            if conn_key not in self.connection_pool:
                self.connection_pool[conn_key] = http.client.HTTPConnection(host, port, timeout=timeout)

            user_agent = random.choice(self.user_agents)
            headers = headers or {}
            headers["Host"] = conn_key
            headers["User-Agent"] = user_agent

            if auth:
                auth_header = self._generate_auth_header(auth)
                headers.update(auth_header)

            conn = self.connection_pool[conn_key]
            body = None
            if json_data:
                data = json.dumps(json_data)
                headers["Content-Type"] = "application/json"
                body = data.encode()

            if data:
                headers["Content-Length"] = str(len(data))
                body = data.encode()

            try:
                if self.proxy:
                    conn.set_tunnel(host, port, headers={'Host': url})
                    conn.connect(self.proxy['host'], self.proxy['port'])

                conn.request(method, path + '?' + query, body=body, headers=headers)
                response = conn.getresponse()

                if 300 <= response.status < 400:
                    location = response.getheader("Location")
                    if location:
                        if not location.startswith(('http://', 'https://')):
                            location = urljoin(url, location)
                        url = location
                        redirect_count += 1
                        continue

                response_text = self._decode_response(response)

                if response.status >= 400:
                    raise HttpException(response.status, response.reason)

                content_type = response.headers.get("Content-Type", "")
                response_obj = Response(response.status, response_text, response.headers, content_type)
                return response_obj
            except HttpException:
                raise 
            except Exception as e:
                self.logger.error(f"Request error: {type(e).__name__}: {e}")
                return None
            finally:
                conn.close()

    def _decode_response(self, response: http.client.HTTPResponse) -> str:
        content_encoding = response.headers.get('Content-Encoding', '')
        if 'gzip' in content_encoding:
            return gzip.decompress(response.read()).decode()
        elif 'deflate' in content_encoding:
            return zlib.decompress(response.read(), -zlib.MAX_WBITS).decode()
        else:
            return response.read().decode()

    def _generate_cache_key(self, method: str, url: str, headers: Optional[Dict[str, str]], params: Optional[Dict[str, Union[str, int]]], data: Optional[str], 
                            json_data: Optional[Dict[str, Any]]) -> str:
        cache_key = hashlib.sha1()
        cache_key.update(method.encode())
        cache_key.update(url.encode())
        cache_key.update(json.dumps(headers).encode())
        cache_key.update(json.dumps(params).encode() if params else b'')
        cache_key.update(data.encode() if data else b'')
        cache_key.update(json.dumps(json_data).encode() if json_data else b'')
        return cache_key.hexdigest()

    def _cache_response(self, cache_key: str, response: Response) -> None:
        if len(self.cache) >= self.response_cache_max_size:
            self.cache.popitem(last=False)
        self.cache[cache_key] = response

    def _generate_auth_header(self, auth: Dict[str, str]) -> Dict[str, str]:
        if auth['type'] == 'basic':
            user_pass = f"{auth['username']}:{auth['password']}"
            basic_auth = base64.b64encode(user_pass.encode()).decode()
            return {"Authorization": f"Basic {basic_auth}"}


import importlib.util

class SessionConfigSettings:
    def __init__(self) -> None:
        self.settings_module_path: str = "./settings.py"

    def _fetch_sessions_config(self):
        try:
            settings = self._load_settings_module()
            sessions_config = getattr(settings, 'SESSIONS_CONFIG', None)

            session_backend = getattr(settings, 'SESSION_BACKEND', None)
            sessions_secret_key = getattr(settings, 'SESSION_SECRET_KEY', None)
            sessions_max_age = getattr(settings, 'SESSION_MAX_AGE', 1800)
            sessions_lifetime = getattr(settings, 'SESSION_LIFETIME', 30)
            sessions_secure = getattr(settings, 'SESSION_SECURE', True)
            sessions_httponly = getattr(settings, 'SESSION_HTTPONLY', True)
            sessions_samesite = getattr(settings, 'SESSION_SAMESITE', 'Lax')
            sessions_cookie_name = getattr(settings, 'SESSION_COOKIE_NAME', '_session_id')

            if sessions_config is None and session_backend is None:
                raise KeyError("Session backend cannot be none either use memory or cookie.")
            if sessions_config is None and sessions_secret_key is None:
                raise KeyError("Session secret_key can't be none! Use a strong key.")
            
            var_session_config = [{
                'backend': session_backend,
                'secret_key': sessions_secret_key,
                'max_age': sessions_max_age,
                'ifetime': sessions_lifetime,
                'secure': sessions_secure,
                'httponly': sessions_httponly,
                'samesite': sessions_samesite,
                'cookie_name': sessions_cookie_name
            }]

            if sessions_config is not None and var_session_config[0]['backend'] is not None:
                raise ValueError("Cannot use both session_config and config_variables simultaneously.")

            if sessions_config is None:
                return var_session_config

            return sessions_config

        except (FileNotFoundError, AttributeError) as e:
            raise ImportError(f"Error loading settings.py: {e}")

    def _load_settings_module(self):
        spec = importlib.util.spec_from_file_location("settings", self.settings_module_path)
        settings = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings)
        return settings

    def _set_settings_data(self, session_config):
        session_lifetime = session_config.get('session_lifetime', 30)
        secret_key = session_config.get('secret_key', )
        max_age = session_config.get('max_age', 1800)
        secure = session_config.get('secure', True)
        httponly = session_config.get('httponly', True)
        samesite = session_config.get('samesite', 'Lax')
        cookie_name = session_config.get('cookie_name', '_sessionid')

        return {
            'max_age': max_age,
            'session_lifetime': session_lifetime,
            'secure': secure,
            'httponly': httponly,
            'samesite': samesite,
            'cookie_name': cookie_name,
            'secret_key': secret_key
        }
    
    def _set_memory_settings_data(self, config):
        session_lifetime = config.get('session_lifetime', 30)
        secret_key = config.get('secret_key', "mysecret_key")

        return { 'session_lifetime': session_lifetime, 'secret_key': secret_key}

    def fetch(self):
        sessions_config = self._fetch_sessions_config()

        backend_types = set()
        for session_config in sessions_config:
            backend = session_config.get('backend')
            backend_types.add(backend)
            secret_key = session_config.get('secret_key')
            if not isinstance(secret_key, str):
                raise ValueError("The 'secret_key' must be a string.")
            if backend == 'cookie':
                return self._set_settings_data(session_config)
            elif backend == 'memory':
                return self._set_memory_settings_data(session_config)

        if len(backend_types) > 1:
            raise ValueError("Multiple session backends detected. Please use only one type of session backend (either cookies or memory-based sessions), not a mix of both.")


from .converter import Converter as Converter

from . import _routing  as routing

from ..settings.core import( 
    BaseSettings as BaseSettings,
    fetchSettingsMiddleware as fetchSettingsMiddleware,
    StageHandler as StageHandler
)


import re

from aquilify.utils.functional import SimpleLazyObject

ESCAPE_MAPPINGS = {
    "A": None,
    "b": None,
    "B": None,
    "d": "0",
    "D": "x",
    "s": " ",
    "S": "x",
    "w": "x",
    "W": "!",
    "Z": None,
}


class Choice(list):
    """Represent multiple possibilities at this point in a pattern string."""


class Group(list):
    """Represent a capturing group in the pattern string."""


class NonCapture(list):
    """Represent a non-capturing group in the pattern string."""


def normalize(pattern):
    result = []
    non_capturing_groups = []
    consume_next = True
    pattern_iter = next_char(iter(pattern))
    num_args = 0
    
    try:
        ch, escaped = next(pattern_iter)
    except StopIteration:
        return [("", [])]

    try:
        while True:
            if escaped:
                result.append(ch)
            elif ch == ".":
                result.append(".")
            elif ch == "|":
                # FIXME: One day we'll should do this, but not in 1.0.
                raise NotImplementedError("Awaiting Implementation")
            elif ch == "^":
                pass
            elif ch == "$":
                break
            elif ch == ")":
                start = non_capturing_groups.pop()
                inner = NonCapture(result[start:])
                result = result[:start] + [inner]
            elif ch == "[":
                ch, escaped = next(pattern_iter)
                result.append(ch)
                ch, escaped = next(pattern_iter)
                while escaped or ch != "]":
                    ch, escaped = next(pattern_iter)
            elif ch == "(":
                ch, escaped = next(pattern_iter)
                if ch != "?" or escaped:
                    name = "_%d" % num_args
                    num_args += 1
                    result.append(Group((("%%(%s)s" % name), name)))
                    walk_to_end(ch, pattern_iter)
                else:
                    ch, escaped = next(pattern_iter)
                    if ch in "!=<":
                        walk_to_end(ch, pattern_iter)
                    elif ch == ":":
                        non_capturing_groups.append(len(result))
                    elif ch != "P":
                        raise ValueError("Non-reversible reg-exp portion: '(?%s'" % ch)
                    else:
                        ch, escaped = next(pattern_iter)
                        if ch not in ("<", "="):
                            raise ValueError(
                                "Non-reversible reg-exp portion: '(?P%s'" % ch
                            )
                        if ch == "<":
                            terminal_char = ">"
                        else:
                            terminal_char = ")"
                        name = []
                        ch, escaped = next(pattern_iter)
                        while ch != terminal_char:
                            name.append(ch)
                            ch, escaped = next(pattern_iter)
                        param = "".join(name)
                        if terminal_char != ")":
                            result.append(Group((("%%(%s)s" % param), param)))
                            walk_to_end(ch, pattern_iter)
                        else:
                            result.append(Group((("%%(%s)s" % param), None)))
            elif ch in "*?+{":
                count, ch = get_quantifier(ch, pattern_iter)
                if ch:
                    consume_next = False

                if count == 0:
                    if contains(result[-1], Group):
                        result[-1] = Choice([None, result[-1]])
                    else:
                        result.pop()
                elif count > 1:
                    result.extend([result[-1]] * (count - 1))
            else:
                result.append(ch)

            if consume_next:
                ch, escaped = next(pattern_iter)
            consume_next = True
    except StopIteration:
        pass
    except NotImplementedError:
        return [("", [])]

    return list(zip(*flatten_result(result)))


def next_char(input_iter):
    for ch in input_iter:
        if ch != "\\":
            yield ch, False
            continue
        ch = next(input_iter)
        representative = ESCAPE_MAPPINGS.get(ch, ch)
        if representative is None:
            continue
        yield representative, True


def walk_to_end(ch, input_iter):
    if ch == "(":
        nesting = 1
    else:
        nesting = 0
    for ch, escaped in input_iter:
        if escaped:
            continue
        elif ch == "(":
            nesting += 1
        elif ch == ")":
            if not nesting:
                return
            nesting -= 1


def get_quantifier(ch, input_iter):
    if ch in "*?+":
        try:
            ch2, escaped = next(input_iter)
        except StopIteration:
            ch2 = None
        if ch2 == "?":
            ch2 = None
        if ch == "+":
            return 1, ch2
        return 0, ch2

    quant = []
    while ch != "}":
        ch, escaped = next(input_iter)
        quant.append(ch)
    quant = quant[:-1]
    values = "".join(quant).split(",")

    try:
        ch, escaped = next(input_iter)
    except StopIteration:
        ch = None
    if ch == "?":
        ch = None
    return int(values[0]), ch


def contains(source, inst):
    if isinstance(source, inst):
        return True
    if isinstance(source, NonCapture):
        for elt in source:
            if contains(elt, inst):
                return True
    return False


def flatten_result(source):
    if source is None:
        return [""], [[]]
    if isinstance(source, Group):
        if source[1] is None:
            params = []
        else:
            params = [source[1]]
        return [source[0]], [params]
    result = [""]
    result_args = [[]]
    pos = last = 0
    for pos, elt in enumerate(source):
        if isinstance(elt, str):
            continue
        piece = "".join(source[last:pos])
        if isinstance(elt, Group):
            piece += elt[0]
            param = elt[1]
        else:
            param = None
        last = pos + 1
        for i in range(len(result)):
            result[i] += piece
            if param:
                result_args[i].append(param)
        if isinstance(elt, (Choice, NonCapture)):
            if isinstance(elt, NonCapture):
                elt = [elt]
            inner_result, inner_args = [], []
            for item in elt:
                res, args = flatten_result(item)
                inner_result.extend(res)
                inner_args.extend(args)
            new_result = []
            new_args = []
            for item, args in zip(result, result_args):
                for i_item, i_args in zip(inner_result, inner_args):
                    new_result.append(item + i_item)
                    new_args.append(args[:] + i_args)
            result = new_result
            result_args = new_args
    if pos >= last:
        piece = "".join(source[last:])
        for i in range(len(result)):
            result[i] += piece
    return result, result_args


def _lazy_re_compile(regex, flags=0):

    def _compile():
        # Compile the regex if it was not passed pre-compiled.
        if isinstance(regex, (str, bytes)):
            return re.compile(regex, flags)
        else:
            assert not flags, "flags must be empty if regex is passed pre-compiled"
            return regex

    return SimpleLazyObject(_compile)



from jinja2 import nodes
from jinja2.ext import Extension
from urllib.parse import urljoin, urlencode

from ...settings import settings

class URLConstructor(Extension):
    tags = {'static', 'redirect', 'media'}

    def __init__(self, environment):
        super().__init__(environment)
        self.static_base_url = getattr(settings, 'STATIC_URL', None)
        self.media_url = settings.MEDIA_URL
        self.static_multi_url = getattr(settings, 'STATIC_MULTI_URLS', [])

    def parse(self, parser):
        tag_name = parser.stream.current.value
        if tag_name in self.tags:
            return self._parse_tag(parser, tag_name)
        else:
            self._raise_invalid_tag_error(tag_name)

    def _parse_tag(self, parser, tag_name):
        line_number = parser.stream.expect(f'name:{tag_name}').lineno
        if tag_name == 'static':
            return self._parse_static_tag(parser, line_number)
        elif tag_name == 'redirect':
            return self._parse_redirect_tag(parser, line_number)
        elif tag_name == 'media':
            return self._parse_media_tag(parser, line_number)

    def _parse_static_tag(self, parser, line_number):
        filename_expr = parser.parse_expression()
        if isinstance(filename_expr, nodes.Const) and filename_expr.value.startswith('multi:'):
            key = filename_expr.value.split(':', 1)[-1].strip()
            return self._create_call_block('_build_multi_static_url', [nodes.Const(key)], line_number)
        else:
            return self._create_call_block('_build_static_url', [filename_expr], line_number)

    def _parse_media_tag(self, parser, line_number):
        filename_expr = parser.parse_expression()
        return self._create_call_block('_build_media_url', [filename_expr], line_number)

    def _parse_redirect_tag(self, parser, line_number):
        url_expr = parser.parse_expression()
        args_expr = parser.parse_expression()
        return self._create_redirect_call_block('_build_redirect_url', [url_expr, args_expr], line_number)

    def _create_call_block(self, method_name, expr_list, line_number):
        return nodes.CallBlock(
            self.call_method(method_name, expr_list),
            [], [], []
        ).set_lineno(line_number)

    def _create_redirect_call_block(self, method_name, expr_list, line_number):
        return nodes.CallBlock(
            self.call_method(method_name, expr_list),
            [], [], []
        ).set_lineno(line_number)

    def _build_static_url(self, filename, caller=None):
        return self._build_url(filename, self.static_base_url, "STATIC_URL")

    def _build_multi_static_url(self, key, caller=None):
        return self._build_multi_url(key, self.static_multi_url, "STATIC_MULTI_URL")

    def _build_media_url(self, filename, caller=None):
        return self._build_url(filename, self.media_url, "MEDIA_URL")

    def _build_redirect_url(self, location, args={}, caller=None):
        self._check_string_type(location, "Invalid redirect location. It must be a string.")
        if args:
            location = urljoin(location, f'?{urlencode(args)}')
        return location

    def _build_url(self, filename, base_url, config_name):
        self._check_string_type(filename, f"Invalid filename. It must be a string for {config_name}.")
        if base_url:
            return urljoin(base_url, filename)
        else:
            raise ValueError(f"No valid URL found in {config_name} configuration.")

    def _build_multi_url(self, key, multi_config, config_name):
        self._check_string_type(key, f"Invalid key. It must be a string for {config_name}.")
        for url_config in multi_config:
            if key in url_config:
                return url_config[key]
        raise ValueError(f"URL for '{key}' not found in {config_name}")

    def _raise_invalid_tag_error(self, tag):
        raise ValueError(f"Invalid tag '{tag}'")

    def _check_string_type(self, value, error_message):
        if not isinstance(value, str):
            raise ValueError(error_message)



import re

class UserAgentParser:
    def __init__(self, user_agent_string: str) -> None:
        self.user_agent_string: str = user_agent_string
        self.browser: str = ''
        self.browser_version: str = ''
        self.browser_engine: str = ''
        self.os: str = ''
        self.os_version: str = ''
        self.device: str = ''
        self.is_mobile: bool = False
        self.language: str = ''
        self.platform: str = ''
        self.is_bot_or_crawler: bool = False
        self.screen_resolution: str = ''
        self.viewport_size: str = ''
        self.js_enabled: bool = False
        self.referer: str = ''
        self.timezone: str = ''
        self._parse_user_agent()

    def _parse_user_agent(self) -> None:
        self.browser, self.browser_version = self._get_browser_info()
        self.browser_engine = self._get_browser_engine()
        self.os, self.os_version = self._get_os_info()
        self.device = self._get_device_info()
        self.is_mobile = self._check_mobile()
        self.language = self._get_language()
        self.platform = self._get_platform()
        self.is_bot_or_crawler = self._check_bot_or_crawler()
        self.screen_resolution = self._get_screen_resolution()
        self.viewport_size = self._get_viewport_size()
        self.js_enabled = self._check_javascript_enabled()
        self.referer = self._get_referer()
        self.timezone = self._get_timezone()

    def _get_browser_info(self):
        browsers = {
            'Opera': r'Opera\/([0-9.]+)',
            'Firefox': r'Firefox\/([0-9.]+)',
            'Edge': r'(?:Edg(?:e)?)\/([0-9.]+)',
            'Chrome': r'Chrome\/([0-9.]+)',
            'Safari': r'Safari\/([0-9.]+)',
            'IE': r'MSIE ([0-9.]+)|rv:([0-9.]+)'
        }
        
        browser_priority = ['Edge', 'Chrome', 'Firefox', 'Safari', 'Opera', 'IE']

        for browser in browser_priority:
            if browser in browsers:
                pattern = browsers[browser]
                match = re.search(pattern, self.user_agent_string)
                if match:
                    if browser == 'Edge':
                        version = match.group(1) or ''
                    else:
                        version = match.group(1) or match.group(2) or ''
                    return browser, version
        
        return 'Unknown', ''
    
    def _get_browser_engine(self):
        browser_engines = {
            'Blink': 'Blink',
            'WebKit': 'WebKit',
            'Gecko': 'Gecko',
            'Trident': 'Trident'
        }
        for engine, pattern in browser_engines.items():
            if pattern in self.user_agent_string:
                return engine
        return 'Unknown'

    def _get_os_info(self):
        operating_systems = {
            'Windows': r'Windows NT ([0-9.]+)',
            'Android': r'Android ([0-9.]+)',
            'Linux': r'Linux ([0-9_]+)',
            'iOS': r'OS ([0-9_]+) like Mac',
            'Mac': r'Mac OS X '
        }
        for os, pattern in operating_systems.items():
            match = re.search(pattern, self.user_agent_string)
            if match:
                version = match.group(1).replace('_', '.') if match.group(1) else ''
                return os, version
        return 'Unknown', ''

    def _get_device_info(self):
        devices = {
            'iPhone': r'iPhone(?:\sSimulator)?',
            'iPad': r'iPad(?:\sSimulator)?',
            'Mobile': r'Mobile',
            'Tablet': r'Tablet',
            'Desktop': r'Windows|Macintosh|Linux'
        }
        for device, pattern in devices.items():
            if re.search(pattern, self.user_agent_string):
                return device
        return 'Unknown'

    def _check_mobile(self):
        return 'Mobile' in self.user_agent_string

    def _get_language(self):
        match = re.search(r'(?<=\b(?:language=))(.*?)(?=[;|$])', self.user_agent_string)
        return match.group(1) if match else 'Unknown'

    def _get_platform(self):
        platforms = {
            'Windows': 'Windows',
            'Linux': 'Linux',
            'Mac': 'Macintosh'
        }
        for platform, pattern in platforms.items():
            if pattern in self.user_agent_string:
                return platform
        return 'Unknown'

    def _check_bot_or_crawler(self):
        bot_patterns = [
            'bot',
            'crawler',
            'spider',
            'googlebot',
            'bingbot',
            'slurp',
            'duckduckbot',
            'yandexbot'
        ]
        for bot_pattern in bot_patterns:
            if re.search(bot_pattern, self.user_agent_string, re.IGNORECASE):
                return True
        return False

    def _get_screen_resolution(self):
        match = re.search(r'(?<=\b(?:Screen: ))([0-9]+x[0-9]+)', self.user_agent_string)
        return match.group(1) if match else 'Unknown'

    def _get_viewport_size(self):
        match = re.search(r'(?<=\b(?:Viewport: ))([0-9]+x[0-9]+)', self.user_agent_string)
        return match.group(1) if match else 'Unknown'

    def _check_javascript_enabled(self):
        return 'JS' in self.user_agent_string

    def _get_referer(self):
        match = re.search(r'(?<=\b(?:Referer: ))(.*?)(?=[;|$])', self.user_agent_string)
        return match.group(1) if match else 'Unknown'

    def _get_timezone(self):
        match = re.search(r'(?<=\b(?:Timezone: ))(.*?)(?=[;|$])', self.user_agent_string)
        return match.group(1) if match else 'Unknown'
    
    def __str__(self) -> str:
        return str(self.user_agent_string)
    
    def __repr__(self) -> str:
        return f"UserAgentParser({self.user_agent_string})"

    def to_dict(self) -> dict:
        return {
            'user_agent_string': self.user_agent_string,
            'browser': self.browser,
            'browser_version': self.browser_version,
            'browser_engine': self.browser_engine,
            'os': self.os,
            'os_version': self.os_version,
            'device': self.device,
            'is_mobile': self.is_mobile,
            'language': self.language,
            'platform': self.platform,
            'is_bot_or_crawler': self.is_bot_or_crawler,
            'screen_resolution': self.screen_resolution,
            'viewport_size': self.viewport_size,
            'js_enabled': self.js_enabled,
            'referer': self.referer,
            'timezone': self.timezone
        }



import os
import base64
import json

from typing import Dict, Any
from datetime import datetime

CACHE_DIR = ".aquilify/cache/session/memory/"
CONFIG_FILE = ".aquilify/cache/session/config.json"
AQUILIFY_CONFIG_FILE = ".aquilify/config.json"

class InMemorySessionStorage:
    def __init__(self, session_id: str, secret_key: str, data: Dict[str, Any] = None) -> None:
        self._data: Dict[str, Any] = data if data else {}
        self._session_id: str = session_id
        self._created_at: datetime = datetime.now()
        self._updated_at: datetime = self._created_at
        self._secret_key = secret_key.encode()
        self._cache_file = os.path.join(CACHE_DIR, f"{self._session_id}.cache")

        os.makedirs(CACHE_DIR, exist_ok=True)
        os.makedirs(os.path.dirname(CONFIG_FILE), exist_ok=True)

        self._update_config()

        self._load_from_cache()
        self._update_aquilify_config()

    def _encrypt(self, data: str) -> str:
        encrypted = []
        key = self._secret_key
        for i in range(len(data)):
            encrypted.append(chr(ord(data[i]) ^ key[i % len(key)]))
        return base64.urlsafe_b64encode("".join(encrypted).encode()).decode()

    def _decrypt(self, encrypted_data: str) -> str:
        decrypted = []
        key = self._secret_key
        decoded_data = base64.urlsafe_b64decode(encrypted_data.encode()).decode()
        for i in range(len(decoded_data)):
            decrypted.append(chr(ord(decoded_data[i]) ^ key[i % len(key)]))
        return "".join(decrypted)
    
    def _load_from_cache(self) -> None:
        try:
            if os.path.exists(self._cache_file):
                with open(self._cache_file, "r") as file:
                    encrypted_data = file.read()
                    decrypted_data = self._decrypt(encrypted_data)
                    self._data = self._deserialize(decrypted_data)
        except Exception as e:
            print(f"Error loading cache: {e}")

    def _save_to_cache(self) -> None:
        try:
            with open(self._cache_file, "w") as file:
                encrypted_data = self._encrypt(self._serialize(self._data))
                file.write(encrypted_data)
        except Exception as e:
            print(f"Error saving cache: {e}")

    def _update_config(self) -> None:
        try:
            with open(CONFIG_FILE, "r") as file:
                config_data = json.load(file)
        except FileNotFoundError:
            config_data = {"$session": []}

        time_stamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        session_entry = {
            "session_id": self._session_id,
            "backend": "memory",
            "cache": True,
            "encryption": "base64",
            "timestamp": time_stamp
        }

        session_ids = [entry['session_id'] for entry in config_data['$session']]
        if self._session_id not in session_ids:
            config_data["$session"].append(session_entry)

        with open(CONFIG_FILE, "w") as file:
            json.dump(config_data, file, indent=4)

    def _update_aquilify_config(self) -> None:
        session_entry = {
            "backend": "memory",
            "cache": True,
            "encryption": "base64",
            "$path": [os.path.join(CACHE_DIR, "config.json")]
        }

        if os.path.exists(AQUILIFY_CONFIG_FILE):
            with open(AQUILIFY_CONFIG_FILE, "r") as aquilify_file:
                aquilify_config = json.load(aquilify_file)
        else:
            aquilify_config = {}

        session_exists = False
        if "$session" in aquilify_config:
            for entry in aquilify_config["$session"]:
                if entry.get("backend") == "memory":
                    session_exists = True
                    break

        if not session_exists:
            existing_compression = aquilify_config.get("$compression", [])

            if "$session" not in aquilify_config:
                aquilify_config["$session"] = []
            aquilify_config["$session"].append(session_entry)

            with open(AQUILIFY_CONFIG_FILE, "w") as aquilify_file:
                aquilify_config["$compression"] = existing_compression
                json.dump(aquilify_config, aquilify_file, indent=4)

    def _serialize(self, data: Dict[str, Any]) -> str:
        return json.dumps(data)

    def _deserialize(self, data: str) -> Dict[str, Any]:
        return json.loads(data)

    def __getitem__(self, key: str) -> Any:
        encrypted_data = self._data.get(key)
        if encrypted_data:
            return self._decrypt(encrypted_data)
        return None

    def __setitem__(self, key: str, value: Any) -> None:
        encrypted_value = self._encrypt(str(value))
        self._data[key] = encrypted_value
        self._save_to_cache()
        self._updated_at = datetime.now()

    def __delitem__(self, key: str) -> None:
        if key in self._data:
            del self._data[key]
            self._updated_at = datetime.now()

    def __contains__(self, key: str) -> bool:
        return key in self._data

    def get(self, key, default=None):
        encrypted_data = self._data.get(key)
        if encrypted_data:
            return self._decrypt(encrypted_data)
        return None

    def set(self, key, value):
        encrypted_value = self._encrypt(str(value))
        self._data[key] = encrypted_value
        self._save_to_cache()
        self._updated_at = datetime.now()

    def delete(self, key):
        if key in self._data:
            del self._data[key]
            self._updated_at = datetime.now()

    def clear(self):
        self._data.clear()
        self._updated_at = datetime.now()

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def items(self):
        return self._data.items()

    def update(self, data: Dict[str, Any]):
        for key, value in data.items():
            encrypted_value = self._encrypt(str(value))
            self._data[key] = encrypted_value
            self._save_to_cache()
            self._updated_at = datetime.now()
            
    @property
    def session_id(self):
        return self._session_id

    @property
    def created_at(self):
        return self._created_at

    @property
    def updated_at(self):
        return self._updated_at
    
    def clear_cache(self) -> None:
        try:
            if os.path.exists(self._cache_file):
                os.remove(self._cache_file)
        except Exception as e:
            print(f"Error clearing cache: {e}")


import importlib.util

class HTTPSConfigSettings:
    def __init__(self) -> None:
        self.settings_module_path: str = "./settings.py"

    def _fetch_https_config(self):
        try:
            settings = self._load_settings_module()
            strict = getattr(settings, 'HTTPS_REDIRECT_STRICT', True)
            allowed_host = getattr(settings, 'HTTPS_REDIRECT_ALLOWED_HOST', [])
            
            return {'strict': strict, 'allowed_host': allowed_host}
        
        except (FileNotFoundError, AttributeError) as e:
            raise ImportError(f"Error loading settings.py: {e}")

    def _load_settings_module(self):
        spec = importlib.util.spec_from_file_location("settings", self.settings_module_path)
        settings = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings)
        return settings


    def fetch(self):
        data = self._fetch_https_config()
        return data


import logging

request_logger = logging.getLogger("aquilify.request")

def log_response(
    message,
    *args,
    response=None,
    request=None,
    logger=request_logger,
    level=None,
    exception=None,
):
    if getattr(response, "_has_been_logged", False):
        return

    if level is None:
        if response.status_code >= 500:
            level = "error"
        elif response.status_code >= 400:
            level = "warning"
        else:
            level = "info"

    getattr(logger, level)(
        message,
        *args,
        extra={
            "status_code": response.status_code,
            "request": request,
        },
        exc_info=exception,
    )
    response._has_been_logged = True